"""
   Wrapper around the job class to build a workflow element (production step + job)
"""

__RCSID__ = "$Id$"

# generic imports
from copy import deepcopy

# DIRAC imports
import DIRAC
from CTADIRAC.Interfaces.API.EvnDispSingJob import EvnDispSingJob
from CTADIRAC.ProductionSystem.Client.WorkflowElement import WorkflowElement


class EvnDispElement(WorkflowElement):
    """Composite class for workflow element (production step + job)"""

    #############################################################################

    def __init__(self, parent_prod_step):
        """Constructor"""
        WorkflowElement.__init__(self, parent_prod_step)
        self.job = EvnDispSingJob(cpuTime=259200.0)
        self.prod_step.Type = "DataReprocessing"

    def build_job_attributes(self, user_job):
        """Set job attributes, with some limitations for file-specific arguments"""
        for key, value in user_job.items():
            if key == "version":
                if value is None:
                    DIRAC.gLogger.error(
                        "Unknown software version for ProdStep %s" % user_job["ID"]
                    )
                    DIRAC.exit(-1)
                else:
                    self.job.version = value
            elif (key == "nsb") & (value is not None):
                self.job.set_output_file_metadata(key, value)
            elif (key == "split") & (value is not None):
                self.job.set_output_file_metadata(key, value)
            else:
                if value is not None:
                    setattr(self.job, key, value)

    def build_element_config(self):
        """Set job and production step attributes specific to the configuration"""
        self.prod_step.Name = "EvnDisp"
        self.prod_step.GroupSize = self.job.group_size
        meta_data = deepcopy(self.prod_step.Inputquery)
        self.job.set_output_metadata(meta_data)
        self.job.set_executable_sequence(debug=False)
        self.prod_step.Body = self.job.workflow.toXML()



"""
   Wrapper around the job class to build a workflow element (production step + job)
"""


__RCSID__ = "$Id$"


# DIRAC imports
import DIRAC
from CTADIRAC.Interfaces.API.MCPipeNSBJob import MCPipeNSBJob
from CTADIRAC.ProductionSystem.Client.WorkflowElement import WorkflowElement


class SimulationElement(WorkflowElement):
    """Composite class for workflow element (production step + job)"""

    #############################################################################

    def __init__(self, parent_prod_step):
        """Constructor"""
        WorkflowElement.__init__(self, parent_prod_step)
        self.job = MCPipeNSBJob()
        self.prod_step.Type = "MCSimulation"

    def build_job_attributes(self, user_job):
        """Set job attributes, with some limitations for corsika arguments which need default values"""
        with_sct = False
        for key, value in user_job.items():
            if key == "version":
                if value is None:
                    DIRAC.gLogger.error(
                        "Unknown software version for ProdStep %s" % user_job["ID"]
                    )
                    DIRAC.exit(-1)
                else:
                    self.job.version = str(value)
            elif key == "site":
                self.job.set_site(value)
            elif key == "particle":
                self.job.set_particle(value)
            elif key == "pointing_dir":
                self.job.set_pointing_dir(value)
            elif key == "sct":
                with_sct = value  # keep the value for later because we also need the version to call set_sct
            elif key == "moon":
                self.job.set_moon(value)
            elif key == "magic":
                self.job.set_magic(value)
            else:
                if value is not None:
                    setattr(self.job, key, value)
        self.job.set_sct_file(with_sct)

    def build_input_data(self, user_job):
        """Simulation Elements do not have input data"""
        self.prod_step.Inputquery = {}

    def build_element_config(self):
        """Set job and production step attributes specific to the configuration"""
        self.prod_step.Name = "MC_Simulation"
        self.job.set_output_metadata()
        self.job.set_executable_sequence(debug=False)
        self.prod_step.Body = self.job.workflow.toXML()

    def build_output_data(self):
        """Build output data from the job metadata and the metadata added on the files"""
        self.prod_step.Outputquery = self.job.output_metadata
        if self.job.moon == "--with-full-moon":
            self.prod_step.Outputquery["nsb"] = {"in": [1, 5, 19]}
        elif self.job.moon == "--with-half-moon":
            self.prod_step.Outputquery["nsb"] = {"in": [1, 5]}
        else:
            self.prod_step.Outputquery["nsb"] = {"in": [1]}

    #############################################################################



"""
   Wrapper around the job class to build a workflow element (production step + job)
"""

__RCSID__ = "$Id$"

# DIRAC imports
import DIRAC
from DIRAC.Interfaces.API.Job import Job
from DIRAC.ProductionSystem.Client.ProductionStep import ProductionStep
from CTADIRAC.Core.Utilities.tool_box import get_dataset_MQ


class WorkflowElement:
    """Composite class for workflow element (production step + job)"""

    #############################################################################

    def __init__(self, parent_prod_step):
        """Constructor"""
        self.job = Job()
        self.prod_step = ProductionStep()
        self.parent_step = parent_prod_step
        self.job.setOutputSandbox(["*Log.txt"])

    def build_job_attributes(self, user_job):
        """Set job attributes"""
        for key, value in user_job.items():
            if key == "version":
                if value is None:
                    DIRAC.gLogger.error(
                        "Unknown software version for ProdStep %s" % user_job["ID"]
                    )
                    DIRAC.exit(-1)
                else:
                    self.job.version = value
            else:
                if value is not None:
                    setattr(self.job, key, value)

    def build_job_common_attributes(self, user_sub):
        """Set job attributes concerning the whole production"""
        if "base_path" in user_sub["Common"]:
            if user_sub["Common"]["base_path"]:
                self.job.base_path = user_sub["Common"]["base_path"]
        if "MCCampaign" in user_sub["Common"]:
            if user_sub["Common"]["MCCampaign"]:
                self.job.MCCampaign = user_sub["Common"]["MCCampaign"]
            else:
                DIRAC.gLogger.error("MCCampaign is mandatory")
                DIRAC.exit(-1)
        if "configuration_id" in user_sub["Common"]:
            if user_sub["Common"]["configuration_id"]:
                self.job.configuration_id = user_sub["Common"]["configuration_id"]
            else:
                DIRAC.gLogger.error("configuration_id is mandatory")
                DIRAC.exit(-1)

    def build_input_data(self, user_job):
        """Build input data from the parent output data or from the dataset metadata"""
        if self.parent_step:
            self.prod_step.Inputquery = self.parent_step.Outputquery
            for key, value in self.job.output_file_metadata.items():
                self.prod_step.Inputquery[key] = value

        else:
            if user_job.get("dataset"):
                self.prod_step.Inputquery = get_dataset_MQ(user_job["dataset"])

            else:
                DIRAC.gLogger.error(
                    "A step without parent step must take a dataset as input"
                )
                DIRAC.exit(-1)

    def build_output_data(self):
        """Build output data from the job metadata and the metadata added on the files"""
        self.prod_step.Outputquery = self.job.output_metadata
        for key, value in self.job.output_file_metadata.items():
            self.prod_step.Outputquery[key] = value


#############################################################################



#!/usr/bin/env python
"""
Launch a production from a workflow description submitted by the user.

Usage:
    cta-prod-submit <name of the production> <YAML file of the production steps description>

Example:
    cta-prod-submit TestProd production_config.yml
"""

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script

Script.parseCommandLine()
import DIRAC
import yaml
import json

from DIRAC.ProductionSystem.Client.ProductionClient import ProductionClient

from CTADIRAC.ProductionSystem.Client.SimulationElement import SimulationElement
from CTADIRAC.ProductionSystem.Client.CtapipeModelingElement import (
    CtapipeModelingElement,
)
from CTADIRAC.ProductionSystem.Client.EvnDispElement import EvnDispElement
from CTADIRAC.ProductionSystem.Client.MergingElement import MergingElement


def check_id(job_list):
    """Check jobID values"""
    for job in job_list:
        if not job.get("ID"):
            DIRAC.gLogger.error("Unknown job ID")
            DIRAC.exit(-1)
        elif not isinstance(job["ID"], int):
            DIRAC.gLogger.error("job ID must be integer")
            DIRAC.exit(-1)
    return True


def sort_by_id(job_list):
    """Sort production steps by ID"""
    return sorted(job_list, key=lambda k: k["ID"])


def check_parents(job_list):
    """Check if parent step is listed before child step"""
    for job in job_list:
        if job.get("parentID"):
            if job["parentID"] > job["ID"]:
                DIRAC.gLogger.error(
                    "A step can only have a parent which ID is inferior to its ID"
                )
                DIRAC.exit(-1)
    return True


def instantiate_workflow_element_from_type(job, parent_prod_step):
    """Instantiate workflow element class based on the job type required"""
    wf_elt = None
    if job["type"].lower() == "mcsimulation":
        wf_elt = SimulationElement(parent_prod_step)
    elif job["type"].lower() == "ctapipeprocessing":
        wf_elt = CtapipeModelingElement(parent_prod_step)
    elif job["type"].lower() == "evndispprocessing":
        wf_elt = EvnDispElement(parent_prod_step)
    elif job["type"].lower() == "merging":
        wf_elt = MergingElement(parent_prod_step)
    else:
        DIRAC.gLogger.error("Unknown job type")
        DIRAC.exit(-1)
    return wf_elt


def find_parent_prod_step(prod_step_list, user_job):
    """Find parent step ID for a given job"""
    parent_prod_step = None
    if "parentID" in user_job:
        if user_job["parentID"] is not None:
            parent_prod_step = prod_step_list[
                user_job["parentID"] - 1
            ]  # Python starts indexing at 0
    return parent_prod_step


def parse_jobs(job_list, prod_sys_client):
    """For each workflow element, build its job and production step"""
    prod_step_list = []
    check_id(job_list["ProdSteps"])
    job_list["ProdSteps"] = sort_by_id(job_list["ProdSteps"])
    check_parents(job_list["ProdSteps"])

    for job in job_list["ProdSteps"]:
        parent_prod_step = find_parent_prod_step(prod_step_list, job)
        workflow_element = instantiate_workflow_element_from_type(job, parent_prod_step)
        workflow_element.build_job_attributes(job)
        workflow_element.build_job_common_attributes(job_list)
        workflow_element.build_input_data(job)
        workflow_element.build_element_config()
        workflow_element.build_output_data()

        DIRAC.gLogger.notice(
            "\tBuilt Production step: %s" % workflow_element.prod_step.Name
        )

        prod_step_list.append(workflow_element.prod_step)
        prod_sys_client.addProductionStep(workflow_element.prod_step)


@Script()
def main():
    arguments = Script.getPositionalArgs()
    if len(arguments) != 2:
        Script.showHelp()

    with open(arguments[1], "r") as stream:
        user_sub = yaml.safe_load(stream)

    ##################################
    # Create the production
    prod_name = "Prod6Test_workflow_%s" % arguments[0]
    DIRAC.gLogger.notice("Building new production: %s" % prod_name)
    prod_sys_client = ProductionClient()

    ##################################
    # Build production steps according to the user submission of the workflow
    parse_jobs(user_sub, prod_sys_client)

    ##################################
    # Get the production description
    prod_description = prod_sys_client.prodDescription
    # Create the production
    DIRAC.gLogger.notice("Creating production.")
    res = prod_sys_client.addProduction(prod_name, json.dumps(prod_description))
    if not res["OK"]:
        DIRAC.gLogger.error(res["Message"])
        DIRAC.exit(-1)

    # Start the production, i.e. instantiate the transformation steps
    res = prod_sys_client.startProduction(prod_name)

    if not res["OK"]:
        DIRAC.gLogger.error(res["Message"])
        DIRAC.exit(-1)

    DIRAC.gLogger.notice("Production %s successfully created" % prod_name)


########################################################
if __name__ == "__main__":
    main()
